<?php

namespace CalculatedFields;

/**
 * Class Fields
 * @package CalculatedFields
 */
class Fields
{

    /**
     * Current state of all fields
     *
     * @var array
     */
    private $fields;

    /**
     * Returns internal state
     *
     * @return array
     */
    public function getFields()
    {
        return $this->fields;
    }

    /**
     * Initialize interal state when being called from
     * ACF save_post handler.
     */
    public function init()
    {
        $this->fields = [];
        $postId = sanitize_text_field($_REQUEST['_acf_post_id']);
        $postedFields = (object)$_REQUEST['acf'];
        foreach ($postedFields as $key => $postedField) {
            $fieldMeta = acf_get_field($key);

            if (!in_array($fieldMeta['type'], ['repeater'])) {
                $this->fields[] = (object)[
                    'name' => $fieldMeta['_name'],
                    'key' => $key,
                    'type' => $fieldMeta['type'],
                    'value' => $postedField,
                    'sortOrder' => $fieldMeta['menu_order'],
                    'formula' => isset($fieldMeta['formula']) ? trim($fieldMeta['formula']) : null,
                ];
            }

            if ($fieldMeta['type'] === 'repeater') {
                $values = [];
                $rowIndex = 0;
                foreach ($postedField as $rowId => $row) {
                    $rowValues = [];
                    $order = 0;
                    foreach ($row as $subKey => $value) {
                        $subFieldMeta = $this->findSubFieldMeta($fieldMeta, $subKey);
                        if (!$subFieldMeta) {
                            continue;
                        }
                        $rowValues[] = (object)[
                            'name' => $subFieldMeta['_name'],
                            'subFieldName' => "{$fieldMeta['name']}_{$rowIndex}_{$subFieldMeta['_name']}",
                            'key' => $subKey,
                            'type' => $subFieldMeta['type'],
                            'value' => $row[$subKey],
                            'sortOrder' => $order++,
                            'formula' => isset($subFieldMeta['formula']) ? trim($subFieldMeta['formula']) : null,
                        ];
                    }
                    $values[] = $rowValues;
                    $rowIndex++;
                }
                $this->fields[] = (object)[
                    'name' => $fieldMeta['_name'],
                    'key' => $key,
                    'type' => $fieldMeta['type'],
                    'value' => $values,
                    'sortOrder' => $fieldMeta['menu_order'],
                    'formula' => isset($fieldMeta['formula']) ? trim($fieldMeta['formula']) : null,
                ];
            }
        }
        $i = 0;
    }

    /**
     * Initialize interal state when being called from ajax
     */
    public function ajaxInit()
    {
        $this->fields = [];
        $postId = sanitize_text_field($_REQUEST['_acf_post_id']);
        $postedFields = json_decode(stripslashes($_REQUEST['acf']));

        foreach ($postedFields as $postedKey => $postedRow) {
            if (strlen(trim($postedKey)) === 0 || $postedKey === '_empty_') {
                continue;
            }
            parse_str($postedKey, $parsedKey);
            $rootKey = key($parsedKey);
            if (!is_array($parsedKey[$rootKey])) {
                continue;
            }
            $key = key($parsedKey[$rootKey]);
            $fieldMeta = acf_get_field($key);

            if (!$fieldMeta) {
                continue;
            }

            if (!in_array($fieldMeta['type'], ['repeater'])) {
                $this->fields[] = (object)[
                    'name' => $fieldMeta['_name'],
                    'key' => $key,
                    'htmlId' => $postedRow->id,
                    'type' => $fieldMeta['type'],
                    'value' => (float)$postedRow->value,
                    'originalValue' => (float)$postedRow->value,
                    'sortOrder' => $fieldMeta['menu_order'],
                    'formula' => isset($fieldMeta['formula']) ? trim($fieldMeta['formula']) : null,
                ];
            }

            if ($fieldMeta['type'] === 'repeater') {
                $rowIndex = key($parsedKey[$rootKey][$key]);
                $subFieldKey = key($parsedKey[$rootKey][$key][$rowIndex]);

                if ($rowIndex === 'acfcloneindex') {
                    continue;
                }

                $fieldIndex = $this->findFieldIndexByKey($key);
                if (is_null($fieldIndex)) {
                    $this->fields[] = (object)[
                        'name' => $fieldMeta['_name'],
                        'key' => $key,
                        'type' => $fieldMeta['type'],
                        'value' => [],
                        'sortOrder' => $fieldMeta['menu_order'],
                        'formula' => isset($fieldMeta['formula']) ? trim($fieldMeta['formula']) : null,
                    ];
                    $fieldIndex = count($this->fields) - 1;
                }

                if (!isset($this->fields[$fieldIndex]->value[$rowIndex])) {
                    $this->fields[$fieldIndex]->value[$rowIndex] = [];
                }
                $subFieldMeta = $this->findSubFieldMeta($fieldMeta, $subFieldKey);
                if (!$subFieldMeta) {
                    continue;
                }
                $this->fields[$fieldIndex]->value[$rowIndex][] = (object)[
                    'name' => $subFieldMeta['_name'],
                    'subFieldName' => "{$fieldMeta['name']}_{$rowIndex}_{$subFieldMeta['_name']}",
                    'key' => $subFieldKey,
                    'htmlId' => $postedRow->id,
                    'type' => $subFieldMeta['type'],
                    'value' => (float)$postedRow->value,
                    'originalValue' => (float)$postedRow->value,
                    'sortOrder' => is_array($this->fields[$fieldIndex]->value[$rowIndex]) ?
                        count($this->fields[$fieldIndex]->value[$rowIndex]):
                        0,
                    'formula' => isset($subFieldMeta['formula']) ? $subFieldMeta['formula'] : null,
                ];
            }
        }
    }

    /**
     * Makes a "best guess" of which fields that are dependencies for all
     * calculated fields. Passed on to fron-end javascript to determine when
     * fields needs to be recalculated
     *
     * @param $fieldsMeta
     * @return array
     */
    public function getDependantFields($fieldsMeta)
    {
        $names = [];
        $depedencies = [];

        foreach ($fieldsMeta as $fieldMeta) {
            if (isset($fieldMeta['formula']) && strlen(trim($fieldMeta['formula'])) > 0) {
                foreach ($names as $name => $obj) {
                    if (strpos($fieldMeta['formula'], $name) !== false) {
                        $depedencies[] = $obj->key;
                    }
                }
            }

            if (isset($fieldMeta['sub_fields'])) {
                foreach ($fieldMeta['sub_fields'] as $subField) {
                    if (isset($subField['formula']) && strlen(trim($subField['formula'])) > 0) {
                        foreach ($names as $name => $obj) {
                            if (strpos($subField['formula'], $name) !== false) {
                                $depedencies[] = $obj->key;
                            }
                        }
                    }
                    $names[$subField['_name']] = (object)[
                        'name' => $subField['_name'],
                        'key' => $subField['key'],
                    ];
                }
            }

            $names[$fieldMeta['_name']] = (object)[
                'name' => $fieldMeta['_name'],
                'key' => $fieldMeta['key'],
            ];
        }
        return array_unique($depedencies);
    }

    /**
     * Find an internal field based on its acf key
     *
     * @param $key
     * @return int|string|null
     */
    private function findFieldIndexByKey($key)
    {
        foreach ($this->fields as $index => $field) {
            if ($field->key === $key) {
                return $index;
            }
        }
        return null;
    }

    /**
     * Find sub field meta data based on parent key and field key
     *
     * @param $parentMeta
     * @param $key
     * @return mixed|null
     */
    private function findSubFieldMeta($parentMeta, $key)
    {
        foreach ($parentMeta['sub_fields'] as $sub_field) {
            if ($sub_field['key'] === $key) {
                return $sub_field;
            }
        }

        return null;
    }
}
